from typing import Callable

from ariadne import ObjectType
from cpl_core.configuration import ConfigurationABC
from cpl_core.dependency_injection import ServiceProviderABC
from cpl_core.type import T
from cpl_discord.service import DiscordBotServiceABC
from cpl_query.extension import List

from bot_api.exception.service_error_code_enum import ServiceErrorCode
from bot_api.exception.service_exception import ServiceException
from bot_api.route.route import Route
from bot_core.configuration.feature_flags_enum import FeatureFlagsEnum
from bot_core.environment_variables import MAINTENANCE
from bot_core.service.permission_service import PermissionService
from bot_data.model.achievement import Achievement
from bot_data.model.auth_role_enum import AuthRoleEnum
from bot_data.model.auth_user import AuthUser
from bot_data.model.auto_role import AutoRole
from bot_data.model.auto_role_rule import AutoRoleRule
from bot_data.model.client import Client
from bot_data.model.known_user import KnownUser
from bot_data.model.level import Level
from bot_data.model.scheduled_event import ScheduledEvent
from bot_data.model.server import Server
from bot_data.model.server_config import ServerConfig
from bot_data.model.short_role_name import ShortRoleName
from bot_data.model.user import User
from bot_data.model.user_joined_game_server import UserJoinedGameServer
from bot_data.model.user_joined_server import UserJoinedServer
from bot_data.model.user_joined_voice_channel import UserJoinedVoiceChannel
from bot_data.model.user_role_enum import UserRoleEnum
from bot_graphql.abc.filter_abc import FilterABC
from bot_graphql.filter.page import Page
from bot_graphql.filter.sort import Sort


class QueryABC(ObjectType):
    __abstract__ = True

    def __init__(self, name: str):
        ObjectType.__init__(self, name)

    def add_collection(self, name: str, get_collection: Callable, filter_type: type = None):
        def wrapper(*args, **kwargs):
            if filter_type is not None and "filter" in kwargs:
                kwargs["filter"] = FilterABC.get_collection_filter(filter_type, kwargs["filter"])
            else:
                kwargs["filter"] = None

            if "page" in kwargs:
                page = Page()
                page.from_dict(kwargs["page"])
                kwargs["page"] = page

            if "sort" in kwargs:
                sort = Sort()
                sort.from_dict(kwargs["sort"])
                kwargs["sort"] = sort

            collection: List = get_collection(*args)
            user = Route.get_user()

            if user == "system" or user.auth_role == AuthRoleEnum.admin:
                return self._resolve_collection(collection, *args, **kwargs)

            return self._resolve_collection(
                collection.where(lambda x: self._can_user_see_element(user, x)), *args, **kwargs
            )

        self.set_field(f"{name}s", wrapper)
        self.set_field(f"{name}Count", lambda *args: wrapper(*args).count())

    def _can_user_see_element(self, user: AuthUser, element: T) -> bool:
        @ServiceProviderABC.inject
        def get_services(services: ServiceProviderABC) -> ServiceProviderABC:
            return services

        @ServiceProviderABC.inject
        def get_config(config: ConfigurationABC) -> ConfigurationABC:
            return config

        services = get_services()
        permissions: PermissionService = services.get_service(PermissionService)
        bot: DiscordBotServiceABC = services.get_service(DiscordBotServiceABC)
        config = get_config()

        if user.auth_role == AuthRoleEnum.admin:
            return True

        for u in user.users:
            guild = bot.get_guild(u.server.discord_id)
            if permissions.is_member_technician(guild.get_member(u.discord_id)):
                return True

        if config.get_configuration(MAINTENANCE):
            return False

        access = False
        if type(element) == Achievement:
            element: Achievement = element
            for u in user.users:
                u: User = u
                guild = bot.get_guild(u.server.discord_id)
                member = guild.get_member(u.discord_id)
                if permissions.is_member_moderator(member) and u.server.id == element.server.id:
                    access = True
                    break

        elif type(element) == AutoRole:
            element: AutoRole = element
            for u in user.users:
                u: User = u
                guild = bot.get_guild(u.server.discord_id)
                member = guild.get_member(u.discord_id)
                if permissions.is_member_moderator(member) and u.server.id == element.server.id:
                    access = True
                    break

        elif type(element) == AutoRoleRule:
            element: AutoRoleRule = element
            for u in user.users:
                u: User = u
                guild = bot.get_guild(u.server.discord_id)
                member = guild.get_member(u.discord_id)
                if permissions.is_member_moderator(member) and u.server.id == element.auto_role.server.id:
                    access = True
                    break

        elif type(element) == Client:
            for u in user.users:
                u: User = u
                if u.server.id == element.server.id:
                    access = True
                    break

        elif type(element) == KnownUser:
            for u in user.users:
                u: User = u
                guild = bot.get_guild(u.server.discord_id)
                member = guild.get_member(u.discord_id)
                if permissions.is_member_moderator(member):
                    access = True
                    break

        elif type(element) == Level:
            for u in user.users:
                u: User = u
                if u.server.id == element.server.id:
                    access = True
                    break

        elif type(element) == Server:
            for u in user.users:
                u: User = u
                if u.server.id == element.id:
                    access = True
                    break

        elif type(element) == User:
            for u in user.users:
                u: User = u
                guild = bot.get_guild(u.server.discord_id)
                member = guild.get_member(u.discord_id)
                if u.id == element.id or permissions.is_member_moderator(member):
                    access = True
                    break

        elif type(element) == UserJoinedServer:
            for u in user.users:
                u: User = u
                guild = bot.get_guild(u.server.discord_id)
                member = guild.get_member(u.discord_id)
                if u.id == element.user.id or permissions.is_member_moderator(member):
                    access = True
                    break

        elif type(element) == UserJoinedVoiceChannel:
            for u in user.users:
                u: User = u
                guild = bot.get_guild(u.server.discord_id)
                member = guild.get_member(u.discord_id)
                if u.id == element.user.id or permissions.is_member_moderator(member):
                    access = True
                    break

        elif type(element) == UserJoinedGameServer:
            for u in user.users:
                u: User = u
                guild = bot.get_guild(u.server.discord_id)
                member = guild.get_member(u.discord_id)
                if u.id == element.user.id or permissions.is_member_moderator(member):
                    access = True
                    break

        elif type(element) == ServerConfig:
            for u in user.users:
                u: User = u
                guild = bot.get_guild(u.server.discord_id)
                member = guild.get_member(u.discord_id)
                if permissions.is_member_technician(member):
                    access = True
                    break

        elif type(element) == ShortRoleName:
            element: ShortRoleName = element
            for u in user.users:
                u: User = u
                guild = bot.get_guild(u.server.discord_id)
                member = guild.get_member(u.discord_id)
                if permissions.is_member_moderator(member) and u.server.id == element.server.id:
                    access = True
                    break

        elif type(element) == ScheduledEvent:
            element: ScheduledEvent = element
            for u in user.users:
                u: User = u
                guild = bot.get_guild(u.server.discord_id)
                member = guild.get_member(u.discord_id)
                if permissions.is_member_moderator(member) and u.server.id == element.server.id:
                    access = True
                    break

        elif type(element) == dict and "key" in element and element["key"] in [e.value for e in FeatureFlagsEnum]:
            for u in user.users:
                u: User = u
                guild = bot.get_guild(u.server.discord_id)
                member = guild.get_member(u.discord_id)
                if permissions.is_member_technician(member):
                    access = True
                    break

        return access

    @ServiceProviderABC.inject
    def _can_user_mutate_data(
        self, server: Server, permission: UserRoleEnum, services: ServiceProviderABC, config: ConfigurationABC
    ):
        permissions: PermissionService = services.get_service(PermissionService)
        bot: DiscordBotServiceABC = services.get_service(DiscordBotServiceABC)

        auth_user = Route.get_user()
        if auth_user == "system" or auth_user.auth_role == AuthRoleEnum.admin:
            return

        member = bot.get_guild(server.discord_id).get_member(
            auth_user.users.where(lambda x: x.server.id == server.id).single().discord_id
        )

        can_edit = lambda x: False
        match permission:
            case UserRoleEnum.moderator:
                can_edit = permissions.is_member_moderator
                if config.get_configuration(MAINTENANCE):
                    can_edit = lambda x: False
            case UserRoleEnum.admin:
                can_edit = permissions.is_member_admin
                if config.get_configuration(MAINTENANCE):
                    can_edit = lambda x: False
            case UserRoleEnum.technician:
                can_edit = permissions.is_member_technician

        if not can_edit(member):
            ex = ServiceException(ServiceErrorCode.Forbidden, f"User not allowed to mutate data")
            raise ex

    # @FilterABC.resolve_filter_annotation
    def _resolve_collection(
        self,
        collection: List,
        *_,
        filter: FilterABC = None,
        page: Page = None,
        sort: Sort = None,
    ):
        if filter is not None:
            collection = filter.filter(collection)

        if sort is not None:
            collection = sort.filter(collection)

        if page is not None:
            collection = page.filter(collection)

        return collection
