import { Component, OnDestroy, OnInit } from "@angular/core";
import { FormBuilder, FormControl, FormGroup } from "@angular/forms";
import { Page } from "../../../../../../models/graphql/filter/page.model";
import { Sort, SortDirection } from "../../../../../../models/graphql/filter/sort.model";
import { Subject, throwError } from "rxjs";
import { Server } from "../../../../../../models/data/server.model";
import { UserDTO } from "../../../../../../models/auth/auth-user.dto";
import { LazyLoadEvent } from "primeng/api";
import { Queries } from "../../../../../../models/graphql/queries.model";
import { AuthService } from "../../../../../../services/auth/auth.service";
import { SpinnerService } from "../../../../../../services/spinner/spinner.service";
import { ToastService } from "../../../../../../services/toast/toast.service";
import { ConfirmationDialogService } from "../../../../../../services/confirmation-dialog/confirmation-dialog.service";
import { TranslateService } from "@ngx-translate/core";
import { DataService } from "../../../../../../services/data/data.service";
import { SidebarService } from "../../../../../../services/sidebar/sidebar.service";
import { ActivatedRoute } from "@angular/router";
import { Query, ScheduledEventListQuery } from "../../../../../../models/graphql/query.model";
import { catchError, debounceTime, takeUntil } from "rxjs/operators";
import { Table } from "primeng/table";
import { ScheduledEventMutationResult } from "../../../../../../models/graphql/result.model";
import { Mutations } from "../../../../../../models/graphql/mutations.model";
import { ComponentWithTable } from "../../../../../../base/component-with-table";
import { EventType, ScheduledEvent, ScheduledEventFilter } from "../../../../../../models/data/scheduled_events.model";

@Component({
  selector: "app-scheduled-events",
  templateUrl: "./scheduled-events.component.html",
  styleUrls: ["./scheduled-events.component.scss"]
})
export class ScheduledEventsComponent extends ComponentWithTable implements OnInit, OnDestroy {
  public scheduledEvents: ScheduledEvent[] = [];
  public loading = true;

  public filterForm!: FormGroup<{
    id: FormControl<number | null>,
    interval: FormControl<string | null>,
    name: FormControl<string | null>,
    description: FormControl<string | null>,
    channelId: FormControl<string | null>,
    startTime: FormControl<string | null>,
    endTime: FormControl<string | null>,
    entityType: FormControl<EventType | null>,
    location: FormControl<string | null>,
  }>;

  public filter: ScheduledEventFilter = {};
  public page: Page = {
    pageSize: undefined,
    pageIndex: undefined
  };
  public sort: Sort = {
    sortColumn: undefined,
    sortDirection: undefined
  };

  public totalRecords: number = 0;

  private unsubscriber = new Subject<void>();
  private server: Server = {};
  public user: UserDTO | null = null;
  public query: string = Queries.scheduledEventWithHistoryQuery;
  public _editableScheduledEvent?: ScheduledEvent = undefined;

  set editableScheduledEvent(event: ScheduledEvent | undefined) {
    if (!event) {
      this.isEditingNew = false;
    } else if (this._editableScheduledEvent && event && this._editableScheduledEvent !== event) {
      this.onRowEditSave(event);
    }
    this._editableScheduledEvent = event;
  }

  get editableScheduledEvent(): ScheduledEvent | undefined {
    return this._editableScheduledEvent;
  }

  public constructor(
    private authService: AuthService,
    private spinner: SpinnerService,
    private toastService: ToastService,
    private confirmDialog: ConfirmationDialogService,
    private fb: FormBuilder,
    private translate: TranslateService,
    private data: DataService,
    private sidebar: SidebarService,
    private route: ActivatedRoute) {
    super("ScheduledEvent", ["id", "interval", "name", "description", "channel_id", "start_time", "end_time", "type", "location"],
      (oldElement: ScheduledEvent, newElement: ScheduledEvent) => {
        return oldElement.name === newElement.name;
      });
  }

  public ngOnInit(): void {
    this.loading = true;
    this.setFilterForm();
    this.data.getServerFromRoute(this.route).then(async server => {
      this.server = server;
      let authUser = await this.authService.getLoggedInUser();
      this.user = authUser?.users?.find(u => u.server == this.server.id) ?? null;
    });
  }

  public ngOnDestroy(): void {
    this.unsubscriber.next();
    this.unsubscriber.complete();
  }

  public loadNextPage(): void {
    this.data.query<ScheduledEventListQuery>(Queries.scheduledEventQuery, {
        serverId: this.server.id, filter: this.filter, page: this.page, sort: this.sort
      },
      (data: Query) => {
        return data.servers[0];
      }
    ).subscribe(data => {
      this.totalRecords = data.scheduledEventCount;
      this.scheduledEvents = data.scheduledEvents;
      this.spinner.hideSpinner();
      this.loading = false;
    });
  }

  public setFilterForm(): void {
    this.filterForm = this.fb.group({
      id: new FormControl<number | null>(null),
      interval: new FormControl<string | null>(null),
      name: new FormControl<string | null>(null),
      description: new FormControl<string | null>(null),
      channelId: new FormControl<string | null>(null),
      startTime: new FormControl<string | null>(null),
      endTime: new FormControl<string | null>(null),
      entityType: new FormControl<EventType | null>(null),
      location: new FormControl<string | null>(null)
    });

    this.filterForm.valueChanges.pipe(
      takeUntil(this.unsubscriber),
      debounceTime(600)
    ).subscribe(changes => {
      if (changes.id) {
        this.filter.id = changes.id;
      } else {
        this.filter.id = undefined;
      }

      if (changes.interval) {
        this.filter.interval = changes.interval;
      } else {
        this.filter.interval = undefined;
      }

      if (changes.name) {
        this.filter.name = changes.name;
      } else {
        this.filter.name = undefined;
      }

      if (changes.description) {
        this.filter.description = changes.description;
      } else {
        this.filter.description = undefined;
      }

      if (changes.channelId) {
        this.filter.channelId = changes.channelId;
      } else {
        this.filter.channelId = undefined;
      }

      if (changes.startTime) {
        this.filter.startTime = changes.startTime;
      } else {
        this.filter.startTime = undefined;
      }

      if (changes.endTime) {
        this.filter.endTime = changes.endTime;
      } else {
        this.filter.endTime = undefined;
      }

      if (changes.entityType) {
        this.filter.entityType = changes.entityType;
      } else {
        this.filter.entityType = undefined;
      }

      if (changes.location) {
        this.filter.location = changes.location;
      } else {
        this.filter.location = undefined;
      }

      if (this.page.pageSize)
        this.page.pageSize = 10;

      if (this.page.pageIndex)
        this.page.pageIndex = 0;

      this.loadNextPage();
    });
  }

  public newScheduledEventTemplate: ScheduledEvent = {
    createdAt: "",
    modifiedAt: ""
  };

  public nextPage(event: LazyLoadEvent): void {
    this.page.pageSize = event.rows ?? 0;
    if (event.first != null && event.rows != null)
      this.page.pageIndex = event.first / event.rows;
    this.sort.sortColumn = event.sortField ?? undefined;
    this.sort.sortDirection = event.sortOrder === 1 ? SortDirection.ASC : event.sortOrder === -1 ? SortDirection.DESC : SortDirection.ASC;

    this.loadNextPage();
  }

  public resetFilters(): void {
    this.filterForm.reset();
  }

  public onRowEditInit(table: Table, event: ScheduledEvent, index: number): void {
    this.editableScheduledEvent = event;
  }

  public override onRowEditSave(newScheduledEvent: ScheduledEvent): void {
    if (this.isEditingNew && JSON.stringify(newScheduledEvent) === JSON.stringify(this.newScheduledEventTemplate)) {
      this.isEditingNew = false;
      return;
    }

    if (this.isEditingNew) {
      this.spinner.showSpinner();
      this.data.mutation<ScheduledEventMutationResult>(Mutations.createScheduledEvent, {
          interval: newScheduledEvent.interval,
          name: newScheduledEvent.name,
          description: newScheduledEvent.description,
          channelId: newScheduledEvent.channelId,
          startTime: newScheduledEvent.startTime,
          endTime: newScheduledEvent.endTime,
          entityType: newScheduledEvent.entityType,
          location: newScheduledEvent.location,
          serverId: this.server.id
        }
      ).pipe(catchError(err => {
        this.isEditingNew = false;
        this.spinner.hideSpinner();
        return throwError(err);
      })).subscribe(result => {
        this.isEditingNew = false;
        this.spinner.hideSpinner();
        this.toastService.success(this.translate.instant("view.server.scheduled_events.message.scheduled_event_create"), this.translate.instant("view.server.scheduled_events.message.scheduled_event_create_d", { name: result.scheduledEvent.createScheduledEvent?.name }));
        this.loadNextPage();
      });
      return;
    }

    this.spinner.showSpinner();
    this.data.mutation<ScheduledEventMutationResult>(Mutations.updateScheduledEvent, {
        id: newScheduledEvent.id,
        interval: newScheduledEvent.interval,
        name: newScheduledEvent.name,
        description: newScheduledEvent.description,
        channelId: newScheduledEvent.channelId,
        startTime: newScheduledEvent.startTime,
        endTime: newScheduledEvent.endTime,
        entityType: newScheduledEvent.entityType,
        location: newScheduledEvent.location,
        serverId: this.server.id
      }
    ).pipe(catchError(err => {
      this.spinner.hideSpinner();
      return throwError(err);
    })).subscribe(_ => {
      this.spinner.hideSpinner();
      this.toastService.success(this.translate.instant("view.server.scheduled_events.message.scheduled_event_update"), this.translate.instant("view.server.scheduled_events.message.scheduled_event_update_d", { name: newScheduledEvent.name }));
      this.loadNextPage();
    });
  }

  public deleteScheduledEvent(ScheduledEvent: ScheduledEvent): void {
    this.confirmDialog.confirmDialog(
      this.translate.instant("view.server.scheduled_events.message.scheduled_event_delete"), this.translate.instant("view.server.scheduled_events.message.scheduled_event_delete_q", { name: ScheduledEvent.name }),
      () => {
        this.spinner.showSpinner();
        this.data.mutation<ScheduledEventMutationResult>(Mutations.deleteScheduledEvent, {
            id: ScheduledEvent.id
          }
        ).pipe(catchError(err => {
          this.spinner.hideSpinner();
          return throwError(err);
        })).subscribe(l => {
          this.spinner.hideSpinner();
          this.toastService.success(this.translate.instant("view.server.scheduled_events.message.scheduled_event_deleted"), this.translate.instant("view.server.scheduled_events.message.scheduled_event_deleted_d", { name: ScheduledEvent.name }));
          this.loadNextPage();
        });
      });
  }

  public addScheduledEvent(): void {
    this.isEditingNew = true;
    this.editableScheduledEvent = JSON.parse(JSON.stringify(this.newScheduledEventTemplate));
  }
}
